@@ -50,8 +50,8 @@
 #endif
 extern int optopt;
 
-#include "zip.h"
 #include "compat.h"
+#include "zip.h"
 
 zip_source_t *source_hole_create(const char *, int flags, zip_error_t *);
 
@@ -402,13 +402,13 @@ name_locate(int argc, char *argv[]) {
 }
 
 static void
-progress_callback(double percentage) {
+progress_callback(zip_t *za, double percentage, void *ud) {
     printf("%.1lf%% done\n", percentage*100);
 }
 
 static int
 print_progress(int argc, char *argv[]) {
-    zip_register_progress_callback(za, progress_callback);
+    zip_register_progress_callback_with_state(za, 0.001, progress_callback, NULL, NULL);
     return 0;
 }
 
@@ -649,6 +649,10 @@ get_compression_method(const char *arg)
         return ZIP_CM_STORE;
     else if (strcmp(arg, "deflate") == 0)
         return ZIP_CM_DEFLATE;
+#if defined(HAVE_LIBBZ2)
+    else if (strcmp(arg, "bzip2") == 0)
+        return ZIP_CM_BZIP2;
+#endif
     else if (strcmp(arg, "unknown") == 0)
         return 100;
     return 0; /* TODO: error handling */
@@ -695,7 +699,13 @@ read_from_file(const char *archive, int flags, zip_error_t *error, zip_uint64_t
     int err;
 
     if (offset == 0 && length == 0) {
-	if ((zaa = zip_open(archive, flags, &err)) == NULL) {
+	if (strcmp(archive, "/dev/stdin") == 0) {
+	    zaa = zip_fdopen(STDIN_FILENO, flags & ~ZIP_CREATE, &err);
+	}
+	else {
+	    zaa = zip_open(archive, flags, &err);
+	}
+	if (zaa == NULL) {
 	    zip_error_set(error, err, errno);
 	    return NULL;
 	}
@@ -722,6 +732,11 @@ read_hole(const char *archive, int flags, zip_error_t *error)
     zip_source_t *src = NULL;
     zip_t *zs = NULL;
 
+    if (strcmp(archive, "/dev/stdin") == 0) {
+	zip_error_set(error, ZIP_ER_OPNOTSUPP, 0);
+	return NULL;
+    }
+
     if ((src = source_hole_create(archive, flags, error)) == NULL
         || (zs = zip_open_from_source(src, flags, error)) == NULL) {
         zip_source_free(src);
@@ -738,6 +753,11 @@ read_to_memory(const char *archive, int flags, zip_error_t *error, zip_source_t
     zip_source_t *src;
     zip_t *zb;
 
+    if (strcmp(archive, "/dev/stdin") == 0) {
+	zip_error_set(error, ZIP_ER_OPNOTSUPP, 0);
+	return NULL;
+    }
+
     if (stat(archive, &st) < 0) {
 	if (errno == ENOENT) {
 	    src = zip_source_buffer_create(NULL, 0, 0, error);
@@ -1023,6 +1043,9 @@ usage(const char *progname, const char *reason)
 	    "\tu\tZIP_FL_UNCHANGED\n");
     fprintf(out, "\nSupported compression methods are:\n"
 	    "\tdefault\n"
+#if defined(HAVE_LIBBZ2)
+	    "\tbzip2\n"
+#endif
 	    "\tdeflate\n"
 	    "\tstore\n");
     fprintf(out, "\nSupported compression methods are:\n"
